import warnings
warnings.simplefilter(action='ignore',category=FutureWarning)
from smurfs.support import *
import matplotlib.pyplot as pl
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import numpy as np
from matplotlib.colors import Colormap

pl.rc('font', family='serif')
pl.rc('xtick', labelsize='x-small')
pl.rc('ytick', labelsize='x-small')

def plot_timeseries(title : str, data : np.ndarray):
    fig : Figure = pl.figure(figsize=(12,4))
    ax : Axes = fig.add_subplot(111)

    ax.plot(data[0],data[1],'x',markersize=3,color='k')
    ax.set_xlabel("Time (days)")
    ax.set_ylabel("Flux")
    ax.set_title(title)

    fig.savefig(f"{title}.pdf")

@timeit
def plotCustom(title : str, data : np.ndarray, **kwargs) -> Figure:
    fig: Figure = pl.figure(figsize=(12,4))
    ax: Axes = fig.add_subplot(111)

    dotList = ['x', 'o']
    lineStyleList = ['-', '--', '-.', ':']

    for name,(data,linestyle) in data.items():
        if linestyle in dotList:
            ax.plot(data[0], data[1],linestyle,label=name,markersize=3,color='k')
        elif linestyle in lineStyleList:
            ax.plot(data[0], data[1], linestyle = linestyle, label=name,linewidth = 1,color='k')
        elif linestyle == '|':
            ax.axvline(x=data,linestyle='--',label=name,linewidth = 1,color='blue')
        elif linestyle == '/':
            ax.axhline(y=data,linestyle='--',label=name,linewidth = 1,color='red')
        else:
            ax.plot(data[0],data[1],label=name)

    ax.set_title(title)
    if 'xLabel' in kwargs.keys():
        ax.set_xlabel(kwargs['xLabel'])

    if 'yLabel' in kwargs.keys():
        ax.set_ylabel(kwargs['yLabel'])
    return fig


def plotMesh(f,t,i,**kwargs):
    fig = pl.figure(figsize=(12,4))
    ax1 = fig.add_subplot(111)
    if "minimumIntensity" in kwargs.keys():
        ax1.pcolormesh(f, t, i,cmap="gnuplot"
                                  ,vmin=kwargs["minimumIntensity"],vmax=np.amax(i))
    else:
        ax1.pcolormesh(f, t, i, cmap="gnuplot")

    if "tMax" in kwargs.keys():
        ax1.set_ylim(0,kwargs["tMax"])
    ax1.set_xlabel(r"Frequency")
    ax1.set_ylabel(r"Time")

    if "frequencyList" in kwargs.keys() and kwargs["frequencyList"] is not None:
        ax2 = ax1.twiny()
        ax3 = ax1.twiny()

        fVal = list(kwargs["frequencyList"].values())
        names = list(kwargs["frequencyList"].keys())
        minX,maxX = min(fVal),max(fVal)

        fVal2 = fVal[::2]
        names2 = names[::2]

        fVal = [x for x in fVal if x not in fVal2]
        names = [x for x in names if x not in names2]

        ax2.set_xticks(fVal)
        ax2.set_xticklabels(names)
        ax2.tick_params(labelrotation=90)

        ax3.set_xticks(fVal2)
        ax3.set_xticklabels(names2)
        ax3.tick_params(pad=20,labelrotation=90)

        ax1.set_xlim(minX*0.95,maxX*1.05)
        ax2.set_xlim(minX* 0.95, maxX * 1.05)
        ax3.set_xlim(minX * 0.95, maxX * 1.05)

    np.save("frequency.npy",f)
    np.save("time.npy", t)
    np.save("intensity.npy", i)

    if not os.path.exists("results"):
        os.mkdir("results")
    try:
        fig.savefig("dynamic_fourier.png")
    except:
        print(term.format("Failed to plot dynamic fourier plot! Maybe only one timerange?",
                          term.Color.YELLOW))



import argparse
from smurfs.analysis import *
from smurfs._version import __version__
from smurfs.support.config import UncertaintyMode, conf

if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument("fileName", help="The filename of the lightcurve. Has to be a text file. Please provide "
                                         "absolute path", type=str)
    parser.add_argument("snr", help="The desired lower bound for the signal to noise ratio for the analysis",
                        type=float)
    parser.add_argument("windowSize", help="The window size used to get the SNR for a given frequency", type=float)
    parser.add_argument("-fr", "--frequencyRange",
                        help="Optional parameter that describes the desired frequency range. "
                             "The application will only check within this parameters if given. "
                             "Per default this parameter is set from 0 to Nyquist frequency."
                             "Please enter it in a form of a tuple (lower,higher).",
                        type=str, default="0,100")
    parser.add_argument("-trs", "--timeBaseSplit",
                        help="Optional parameter that describes the split within the time range."
                             "This can be used to get multiple amplitude spectra for the "
                             "lightcurve, split up within the time Range. A parameter could be "
                             "for example 50, where the parameter would split the lightcurve in"
                             "50 day chunks (last segment will be smaller/equal). Defaults"
                             "to -1, which means no splitting will occur",
                        type=int, default=-1)
    parser.add_argument("-o", "--overlap",
                        help="Optional parameter that describes the overlap of two split ranges. Will only"
                             "work, if -trs is set to a reasonable parameter. For example, if the trs is "
                             "set to 50 and the overlap to 2, the chunks would contain the ranges (0,50),"
                             "(48,98),(96,146) ... Defaults to 0, which means no overlap will take place",
                        type=int, default=0)
    parser.add_argument("-om", "--outputMode", help="Optional parameter that defines what will be exported as dataset. "
                                                    "There are two possibilities: Normal and Full. Normal mode means, that"
                                                    "only the first and last amplitudespectra will be plotted & exported, "
                                                    "full means all amplitudespectra will be plotted & exported. Default is "
                                                    "Normal.",
                        type=str, choices=["Normal", "Full"], default="Normal")
    parser.add_argument("-fm", "--frequencyMarker",
                        help="Optional parameter, that can be passed if frequencies should be "
                             "marked in the dynamic fourier spektrum. Frequencies will be plotted "
                             "on top of the plot. Please use a relative path to the call path "
                             "you are currently in",
                        type=str, default="")
    parser.add_argument("--version", help="Shows version of SMURFS", action='version',
                        version='SMURFS {version}'.format(version=__version__))

    parser.add_argument("-igr", "--ignoreCutoffRatio", help="Optional parameter. If this is set to True, it will ignore"
                                                            "the gap ratio cutoff criterion", action='store_true')

    parser.add_argument("-ssa", "--skipSimilarFrequencies", help="If this parameter is set, the frequencies surrounding"
                                                                 "one frequency are in a very similar range, that area "
                                                                 "will be ignored", action='store_true')

    parser.add_argument("-um", "--uncertaintyMode", help="Optional parameter. Set this parameter to choose the "
                                                         "error determination. Either Montgomery & O'Donoghue (1999),"
                                                         "least square errors (fits) or none", type=str
                        , choices=UncertaintyMode.content(), default=UncertaintyMode.fit.value)

    parser.add_argument("-sc", "--skipCutoff",
                        help="Skips the cutoff check, so similar frequencies don't chancel the run. Be aware that this "
                             "might lead to unknown behaviour",
                        action='store_true')

    args = parser.parse_args()

    if "," in args.fileName:
        files = args.fileName.split(",")
    else:
        files = [args.fileName]

    for file in files:
        conf().uncertaintiesMode = args.uncertaintyMode
        conf().skipSimilarFrequencies = args.skipSimilarFrequencies
        conf().skipCutoff = args.skipCutoff

        fData = args.frequencyRange.split(",")
        frequencyRange = (float(fData[0]), float(fData[1]))

        run(file, args.snr, windowSize=args.windowSize
            , frequencyRange=frequencyRange
            , timeRange=args.timeBaseSplit
            , overlap=args.overlap
            , outputMode=args.outputMode
            , frequencyMarker=args.frequencyMarker
            , ignoreCutoffRatio=args.ignoreCutoffRatio)



class Singleton:
    '''
    Non thread safe helper class to ease implementing singleton patterns.
    This should be used as a decorator to the class that should be a singleton.

    The decorated class can define one '__init__' function that takes only
    the 'self' argument. Other than that, there are no restrictions that apply
    to the decorated class.

    To get the singleton instance, use the 'Instance' method. Trying to use
    '__call__' will result in a 'TypeError' being raised.

    Limitations: The decorated class cannot be inherited from.
    '''
    def __init__(self,decorated):
        self.__decorated = decorated
    def ins(self,*args):
        '''
        Returns the singleton instance. Upon its firsct call, it creates a new
        instance of the decorated class and calls its '__init__' method.
        On all subsequent calls, the already created instance is returned
        '''
        if not hasattr(self,"_instanceList"):
            self._instanceList = {}
        try:
            return self._instanceList[args]
        except KeyError:
            self._instanceList[args] = self.__decorated(args)

            return self._instanceList[args]

    def deleteItem(self,*args):
        if not hasattr(self,"_instanceList"):
            return

        try:
            item =  self._instanceList[args]
        except KeyError:
            return

        del item

    def __call__(self):
        raise TypeError("Singletons must be accessed through 'Instance()'.")
    def __instancecheck__(self,inst):
        return isinstance(inst,self.__decorated)



import os
import time
from stem.util import term

import matplotlib as mpl
if os.environ.get('DISPLAY','') == '':
    print('no display found. Using non-interactive Agg backend')
    mpl.use('Agg')

class cd:
    """
    Directory changer. can change the directory using the 'with' keyword, and returns to the previous path
    after leaving intendation. Example:

    with cd("some/path/to/go"): # changing dir
        foo()
        ...
        bar()
    #back to old dir
    """
    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)

printTimer = False

def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            if printTimer:
                print(term.format('%r  %2.2f ms' % \
                      (method.__name__, (te - ts) * 1000),term.Color.BLUE))
        return result
    return timed

processList = {}
multiProcessing = False

def addProcess(name,p):
    p.start()
    if multiProcessing:
        processList[name]=p
    else:
        p.join()

def waitForProcessesFinished():
    for name,process in processList.items():
        print("Waiting for "+name+" to end")
        process.join()


import numpy as np
from astropy.stats import LombScargle
from scipy.optimize import curve_fit
from scipy.signal.windows import get_window
from typing import Tuple, List
import warnings

from smurfs.files import saveAmpSpectrumAndImage,save_frequency_spacing,getMostCommonStep
from smurfs.support import *
from uncertainties import ufloat
from smurfs.support.config import conf,UncertaintyMode

def calculateSpectralWindow(data:np.ndarray, frequencyBoundary: Tuple[float, float] = (0, 50)) -> np.ndarray:
    """
    Computes the spectral window of a given dataset. Derivation of formula from Asteroseismology (2010)
    :param data: time series dataset
    :param frequencyBoundary: boundaries for spectral window
    :return: FFT of window
    """
    if len(data) != 2:
        raise ValueError("Please move the whole dataset through!")

    t = data[0]
    W = []
    f = np.linspace(frequencyBoundary[0],frequencyBoundary[1] , num=50*(frequencyBoundary[1]-frequencyBoundary[0]))
    for i in f:
        w = (1 / len(t)) * np.sum(np.exp(2 * np.pi * 1j * i * t))
        W.append(w)

    p = np.abs(np.array(W))
    return np.array((f,p))

@timeit
def calculateAmplitudeSpectrum(data: np.ndarray, frequencyBoundary: Tuple[float, float] = (0, 50)) -> np.ndarray:
    """
    This function computes a periodogram using the LombScargle algorithm. There are different implementations available
    and we leave it up to astropy to decide which one should be applied
    :param data: Temporal dataset, an observation by some kind instrument. First axis time, second axis flux
    :param frequencyBoundary: Boundary that should be considered when computing the spectrum
    :return: 2-D Array, containing frequency (first array) and amplitude (second array)
    """
    if frequencyBoundary[0] > frequencyBoundary[1]:
        raise ValueError(b"f_min has to be smaller than f_max!")

    # create object
    ls = LombScargle(data[0], data[1], normalization='psd')

    # if defined max frequency is bigger than the nyquist frequency, cutoff at nyquist
    nFrequency = nyquistFrequency(data)
    if frequencyBoundary[1] > nFrequency:
        print(term.format("Upper Frequencyrange of {0} is higher than Nyquist frequency. Please be aware that "
                          "this can lead to problems with the results!"
                          .format(frequencyBoundary[1], nyquistFrequency),term.Color.YELLOW))

    max_frequency = frequencyBoundary[1]

    # compute Spectrum
    f, p = ls.autopower(minimum_frequency=frequencyBoundary[0], maximum_frequency=max_frequency, samples_per_peak=100)

    # normalization of psd in order to get good amplitudes
    p = np.sqrt(4 / len(data[0])) * np.sqrt(p)

    # removing first item
    p = p[1:]
    f = f[1:]

    removeSectors = conf().removeSector

    for (lower,upper) in removeSectors:
        tmpArray = p[f<lower]
        p = np.append(tmpArray,p[f>upper])

        tmpArray = f[f < lower]
        f = np.append(tmpArray, f[f > upper])
    # restricting values to upper boundary
    p = p[f < frequencyBoundary[1]]
    f = f[f < frequencyBoundary[1]]

    return np.array((f, p))


@timeit
def nyquistFrequency(data: np.ndarray) -> float:
    """
    Computes Nyquist frequency.
    :param data: Dataset, in time domain.
    :return: Nyquist frequency of dataset
    """
    return float(1 / (2 * findMostCommonDiff(data[0])))


def findMostCommonDiff(time: np.ndarray) -> float:
    """
    Finds the most common time difference between two data points.
    :param time:
    :return:
    """
    realDiffX = time[1:len(time)] - time[0:len(time) - 1]
    realDiffX = realDiffX[realDiffX!=0]
    (values, counts) = np.unique(realDiffX, return_counts=True)
    mostCommon = values[np.argmax(counts)]
    return mostCommon


def findMaxPowerFrequency(ampSpectrum: np.ndarray):
    maxY = max(ampSpectrum[1])
    maxX = ampSpectrum[0][abs(ampSpectrum[1] - max(ampSpectrum[1])) < 10 ** -4][0]

    return maxY, maxX


def findAndRemoveMaxFrequency(lightCurve: np.ndarray, ampSpectrum: np.ndarray) -> Tuple[List[float], np.ndarray]:
    """
    Finds the frequency with maximum power in the spectrum, fits it to the dataset and than removes this frequency
    from the initial dataset.
    :param lightCurve: Lightcurve dataset. This is the data where the fit is applied to, and its removed version is
    returned
    :param ampSpectrum: Amplitude spectrum of lightCurve. Should be computed with the same dataset that is provided here
    :return: Fit parameters, Reduced lightcurve
    """
    maxY, maxX = findMaxPowerFrequency(ampSpectrum)
    popt = [-1, -1, -1]
    arr = [maxY, maxX, 0]

    # First fit could provide negative values for amplitude and frequency if the phase is moved by pi. Therefore run
    # again, with phase moved by pi as long as we don't have positive values for freuqency and amplitude
    while popt[0] < 0 or popt[1] < 0:
        try:
            popt, pcov = curve_fit(sin, lightCurve[0], lightCurve[1], p0=arr)
        except RuntimeError:
            print(term.format("Failed to find a good fit for Frequency " + str(maxX) + "c/d", term.Color.RED))
            raise RuntimeError

        # if amp < 0 -> move sin to plus using the phase by adding pi
        if popt[0] < 0 or popt[1] < 0:
            popt[0] = abs(popt[0])
            popt[1] = abs(popt[1])
            popt[2] += -np.pi if popt[2] > np.pi else np.pi
        arr = popt

        while popt[2] < 0:
            popt[2] += 2*np.pi

        while popt[2] > 2*np.pi:
            popt[2] -= 2*np.pi

        retLightCurve = np.array((lightCurve[0], lightCurve[1] - sin(lightCurve[0], *popt)))

    if defines.minimumIntensity is None or defines.minimumIntensity > popt[0]:
        defines.minimumIntensity = popt[0]

    perr = np.sqrt(np.diag(pcov))
    ret = []
    if conf().uncertaintiesMode == UncertaintyMode.fit.value:
        for i in range(0,len(popt)):
            ret.append(ufloat(popt[i],perr[i]))
    elif conf().uncertaintiesMode == UncertaintyMode.montgomery.value:
        #computation of uncertainties with Montgomery & O'Donoghue (1999)
        sigma_m = np.std(lightCurve[1])
        sigma_amp = np.sqrt(2/len(lightCurve[1])) * sigma_m
        sigma_f = np.sqrt(6/len(lightCurve[1])) * (1/(np.pi*max(lightCurve[0])-min(lightCurve[0]))) * sigma_m/popt[0]
        sigma_phi = np.sqrt(2/len(lightCurve[1])) * sigma_m/popt[0]
        ret = [
            ufloat(popt[0],sigma_amp),
            ufloat(popt[1],sigma_f),
            ufloat(popt[2],sigma_phi)
        ]
    elif conf().uncertaintiesMode == UncertaintyMode.none.value:
        for i in popt:
            ret.append(ufloat(i,0))
    else:
        raise ValueError(f"Unknown error computation with {conf().uncertaintiesMode}")


    return ret, retLightCurve


@timeit
def computeSignalToNoise(ampSpectrum: np.ndarray, windowSize: float) -> float:
    """
    Computes signal to noise ratio at frequency f.
    :param ampSpectrum:
    :param windowSize:
    :return:
    """
    y = ampSpectrum[1]
    x = ampSpectrum[0]

    # find minimas adjacent to maxima of y
    lowerMinima, upperMinima = findNextMinimas(y)
    singleStep = np.mean(np.diff(x))
    length_half = int(windowSize / singleStep)

    data = np.append(y[lowerMinima - length_half:lowerMinima], y[upperMinima:upperMinima + length_half])
    meanValue = np.mean(data)
    maxVal = y[abs(y - max(y)) < 10 ** -6][0]

    return float(maxVal / meanValue)


@timeit
def findNextMinimas(yData: np.ndarray) -> Tuple[int, int]:
    index = int(np.where(abs(yData - max(yData)) < 10 ** -6)[0][0])
    minimaFound = False
    counter = 1
    lowerMinima = -1
    upperMinima = -1
    while (minimaFound == False):
        negCounter = index - counter
        posCounter = index + counter
        if negCounter -1 < 0:
            lowerMinima = negCounter
        elif checkMinima(yData, negCounter):
            lowerMinima = negCounter

        if posCounter + 1 >= len(yData):
            upperMinima = posCounter
        elif checkMinima(yData, posCounter):
            upperMinima = posCounter

        if lowerMinima != -1 and upperMinima != -1:
            minimaFound = True
        counter += 1

    return lowerMinima, upperMinima


def checkMinima(yData: np.ndarray, counter: int) -> bool:
    return yData[counter] < yData[counter + 1] and yData[counter] < yData[counter - 1]


def sin(x: np.ndarray, amp: float, f: float, phase: float) -> np.ndarray:
    return amp * np.sin(2 * np.pi * f * x + phase)


def cutoffCriterion(frequencyList: List):
    if len(frequencyList) < similarFrequenciesCount:
        return True
    fList = []

    for i in frequencyList:
        try:
            fList.append(i[0].nominal_value)
        except IndexError:
            try:
                fList.append(i.nominal_value)
            except AttributeError:
                fList.append(i)

    lowerIndex = 0 if len(fList)<similarFrequenciesCount else len(fList)-similarFrequenciesCount
    upperIndex = len(fList)-1
    lastFrequencies = np.array(fList[lowerIndex:upperIndex])
    stdDev = lastFrequencies.std()
    if stdDev < similarityStdDev and not conf():
        print(
            term.format("The last " + str(similarFrequenciesCount) + " where to similar, with a standard deviation of "
                        + str(stdDev), term.Color.RED))
        if conf().skipSimilarFrequencies:
            conf().removeSector.append((lastFrequencies.mean()-10*stdDev,lastFrequencies.mean()+10*stdDev))
            print(term.format(f"Ignoring from {conf().removeSector[-1][0]} to {conf().removeSector[-1][1]}",term.Color.RED))
            return True
        else:
            print(term.format(f"Ending this sector",term.Color.RED))
            return False
    else:
        return True


def prepareSpectrogram(spectrum: np.ndarray, timerange: Tuple[int, int]) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    t = np.linspace(timerange[0], timerange[1],num=5)
    f = spectrum[0]
    i = spectrum[1]

    if defines.maxiLength is None:
        defines.maxiLength = len(i)
    elif defines.maxiLength > len(i):
        i = np.append(i, np.linspace(0, 0, num=defines.maxiLength - len(i)))
    else:
        i = i[0:defines.maxiLength]

    tmpSpectrum = i

    for j in range(0, len(t) - 1):
        i = np.vstack((i, tmpSpectrum))

    return f, t, i


def recursiveFrequencyFinder(data: np.ndarray, snrCriterion: float, windowSize: float, path: str = "",
                             **kwargs):
    """
    Recursively finds all significant frequencies of a given dataset. Takes a lightcurve, transforms it
    using the lomb scargle algorithm, finds the frequency with highest power, fits the frequency to the
    dataset and subtracts it from the dataset. Also prepares the result for the spectrogram plotted at the
    end of the dataset.
    :param data: Input data. Represents a lightcurve. Requires a two dimensional array, with the first
    dimension being the temporal axis, and the second the intensity.
    :param snrCriterion: Cutoff criterion that defines significancy. The code computes the signal to noise
    ratio after each subtraction of the fitted data and checks if the given SNR is bigger than the calculated
    SNR. If this is the case, the algorithm stops.
    :param windowSize: The windowsize of the spectrum. There are also sanity checks going on here, if the
    maximum frequency is higher than the nyquist frequency than the nyquist frequency is used.
    :param path: Path to save stuff to.
    :param kwargs: Commandline parameters.
    :return: List of significant frequencies, initial spectrum, time array, intensity matrix (for spectrogram)
    """
    frequencyList = None
    f, t, i = None, None, None
    print(term.format(f"Nyquist frequency: {nyquistFrequency(data)} c/d", term.Color.CYAN))
    spec_window_saved = False
    try:
        snr = 100
        frequencyList = []
        print(term.format("List of frequencys, amplitudes, phases, S/N", term.Color.CYAN))
        savePath = path + "{0:0=3d}".format(int(data[0][0]))
        savePath += "_{0:0=3d}/".format(int(max(data[0])))

        while (snr > snrCriterion):

            try:
                resNoise = np.mean(data[1])
                frequencyList.append((fit[1], snr, fit[0], fit[2],resNoise))
                saveStuff = True if kwargs['mode'] == 'Full' else False
            except UnboundLocalError:
                saveStuff = True
            amp = calculateAmplitudeSpectrum(data, kwargs['frequencyRange'])
            if not spec_window_saved:
                specWindow = calculateSpectralWindow(data, kwargs['frequencyRange'])
            snr = computeSignalToNoise(amp, windowSize)
            fit, data = findAndRemoveMaxFrequency(data, amp)

            print(term.format(str(fit[1]) + "c/d     " + str(fit[0]) + "     " + str(fit[2]) + "    " + str(snr),
                              term.Color.CYAN))

            amp_spectrum_filename = "amplitude_spectrum_f_" + str(len(frequencyList))
            timeseries_filename = kwargs["name"].split(".")[0] + f"_{len(frequencyList)}" + kwargs["name"].split(".")[1]

            if saveStuff:
                plot_timeseries(timeseries_filename, data)
                np.savetxt(timeseries_filename,data.T)
                saveAmpSpectrumAndImage(amp, savePath, amp_spectrum_filename,amp_spectrum_filename)
                if not spec_window_saved:
                    spec_window_saved = True
                    saveAmpSpectrumAndImage(specWindow, savePath, "","spectral_window_"+ str(len(frequencyList)))
                f, t, i = prepareSpectrogram(amp, (int(data[0][0]), int(np.max(data[0]))))

            if not cutoffCriterion(frequencyList):
                break

        if len(frequencyList) == 0:
            resNoise = np.mean(data[1])
            frequencyList.append((fit[1], snr, fit[0], fit[2], resNoise))

        f_arr = np.array([i[0] for i in frequencyList])

        for roll_index in range(1,len(f_arr)):
            diff = np.abs(f_arr - np.roll(f_arr,roll_index))
            try:
                total_diff_list = np.append(total_diff_list,diff)
            except NameError:
                total_diff_list = diff
        try:
            save_frequency_spacing(total_diff_list,savePath,"frequency_spacing")
        except NameError:
            save_frequency_spacing(f_arr - f_arr,savePath,"frequency_spacing")

        try:
            if kwargs['mode'] == 'Normal':
                pass
                plot_timeseries(timeseries_filename, data)
                np.savetxt(timeseries_filename,data.T)
                saveAmpSpectrumAndImage(amp, savePath, amp_spectrum_filename,amp_spectrum_filename)
        except KeyError:
            pass
    except KeyboardInterrupt:
        print(term.format("Interrupted Run", term.Color.RED))
        defines.dieGracefully = True
    finally:
        conf().removeSector = []
        return frequencyList, f, t, i

def combineDatasets(fList: List[np.ndarray], tList: List[np.ndarray], iList: List[np.ndarray]) -> Tuple[
    np.ndarray, np.ndarray, np.ndarray]:
    if fList == [] or tList == [] or iList == []:
        raise ValueError(term.format("Size of all result lists must be equal!",term.Color.RED))

    for i in tList:
        try:
            t = np.hstack((t,i))
        except UnboundLocalError:
            t = i

    for j in iList:
        try:
            intensity = np.row_stack((intensity,j[:,::]))
        except UnboundLocalError:
            intensity = j[:,::]

    for k in fList:
        if len(k[::]) == intensity.shape[1]:
            f = k[::]
            break

    with warnings.catch_warnings():
        warnings.filterwarnings("error")

        try:
            intensity = np.log10(intensity)
        except RuntimeWarning:
            pass

    return f,t,intensity



from smurfs.files import *
from smurfs.timeseries import *
from smurfs.support import *
import datetime
import re

@timeit
def run(file: str, snrCriterion: float, windowSize: float, **kwargs):
    """
    This function perfomrs a full analysis of a given star. Initially it will split up the dataset according to
    the parameters and than run the recursiveFrequencyFinder, which will return the frequencylist for the given split
    of the original dataset. At the end it will write a results File, containing all splits with their corresponding
    significant (read snr > snrcriterion) frequencies and snrs.
    A results folder will also be created, wherein the first and last amplitude spectrum are stored as well as a plot of
    these.
    :param file: The path to the file where the lightCurve is stored. The lightcurve should contain the time axis in
    its first column and the luminosity in its second column. The code doesn't check against any kind of units, it is
    therefore the responsibility of the user to properly analyse the data afterhand.
    :param snrCriterion: This parameter defines the cutoff criterion for the analysis. It will therefore gather
    frequencies up to the point where its respective SNR is smaller than this snrcriterion.
    :param windowSize: The window size defines the box around the peak from where the noise is calculated. Keep in mind,
    that the window is neither necessarily symmetric nor is the left side and right side of the box starting at the
    peak. To determine the box, it will search for a minimum adjacent to the peak and start the box from this minimum
    on.
    :param kwargs: kwargs defines some other parameters that can be used to steer the code properly. These include:
        - timeRange: The timeRange parameter defines the "box" where the analysis takes place. For example, if the
        dataSet is 450 days long and the timeRange parameter is set to 50, it will split this dataSet in equal size
        chunks, and an independent analysis is performed for each chunk. These can also overlap,
        using the overlap parameter.
        - overlap: The overlap paramater defines the overlap between two chunks, which size is determined by the
        timeRange parameter. Assume the timeRange parameter is set to 50, and the overlap parameter is set to 2. The
        chunks will than represent the dataset from 0->50,48->98,96->146, ...
        - frequencyRange: The frequencyRange parameter defines the frequency range where the dataset is constrained to.
        This will vastly improve speed, but could cut off significant frequencies
        - frequencyMarker: Adds frequency marker to the dynamic fourier plot. If None, no are added.
    """
    print(term.format(f"Running {file}",term.Color.GREEN))
    print(term.format(f"-------------------------------------------", term.Color.GREEN))
    start_time = datetime.datetime.now()
    save_ascii = False
    try:
        if len(file.split(".")) == 1:
            tic_id = re.findall(r"TIC(\d+)",file)[0]
            fileData = download_tic(tic_id)
            save_ascii = True
        else:
            raise IndexError("Use filename!")
    except IndexError:
        fileData = readData(file)

    path = f"results/{file.split('/')[-1].split('.')[0]}_snr_{int(snrCriterion)}_w_{int(windowSize)}"
    if kwargs['timeRange'] != -1:
        path += f"_trs_{int(kwargs['timeRange'])}"
    if kwargs['overlap'] != 0:
        path += f"_o_{int(kwargs['overlap'])}"
    if kwargs['frequencyRange'] != (0, 100):
        path += f"_fr_{int(kwargs['frequencyRange'][0])}_{int(kwargs['frequencyRange'][1])}"

    path += "/"
    createPath(path)

    with cd(path):
        fileData = normalizeData(fileData)

        plot_timeseries(file.split("/")[-1],fileData)

        splitLists = getSplits(fileData,kwargs['timeRange'],kwargs['overlap'],kwargs['ignoreCutoffRatio'])

        try:
            frequencyMarker = readFrequencyMarker(kwargs['frequencyMarker'])
        except FileNotFoundError:
            if not kwargs['frequencyMarker'] == "":
                print(term.format("File {0} was not found. No marker will be added to "
                                  "dynamic fourier plot!".format(kwargs['frequencyMarker']),term.Color.RED))
            frequencyMarker = None
        except ValueError as e:
            print(term.format("Error in frequency Marker file. Error: {0}".format(e),term.Color.RED))
            frequencyMarker = None

        result = {}



        fList = []
        tList = []
        iList = []
        total_f = 0

        for data in splitLists:
            print(term.format("Time base from " + str(int(data[0][0])) + " to " + str(int(max(data[0])))+ " days", term.Color.GREEN) )
            print(term.format("Calculation from "+str(kwargs['frequencyRange'][0])+"c/d to "+str(kwargs['frequencyRange'][1])+"c/d",term.Color.GREEN))

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                try:
                    frequencyList,f,t,i = recursiveFrequencyFinder(data,snrCriterion,windowSize
                                                         ,frequencyRange=kwargs['frequencyRange'],mode=kwargs['outputMode'],name=file)
                except (ValueError,IndexError):
                    print(term.format("Time base from "+ str(int(data[0][0])) + " to " + str(int(max(data[0])))+ "failed "
                             "to perform Lomb-Scargle. Cannot determine Spectrum. Skipping this sector",term.Color.RED))
                    continue

                fList.append(f)
                tList.append(t)
                iList.append(i)

                try:
                    print(term.format(f"Residual noise {abs(frequencyList[-1][4])}", term.Color.GREEN))
                except IndexError:
                    pass

                total_f += len(frequencyList)
                result[(data[0][0], max(data[0]),frequencyList[-1][4])] = frequencyList
                if defines.dieGracefully:
                    break

        f,t,i = combineDatasets(fList,tList,iList)
        if kwargs['timeRange'] == -1:
            tMax = max(t)
        else:
            tMax = max(t) - kwargs['overlap']
        plotMesh(f,t,i,frequencyList = frequencyMarker,minimumIntensity = defines.minimumIntensity,tMax = tMax)


        waitForProcessesFinished()

        if save_ascii:
            np.savetxt(f"TIC_{tic_id}.txt",fileData)

        writeResults("results.txt",result,nyquistFrequency(fileData))

        gap_ratio = getGapRatio(fileData,0,len(fileData[0])-1)

        delta = datetime.datetime.now() - start_time
        minutes = delta.seconds//60
        rest_seconds = delta.seconds%60
        print(term.format(f"Total runtime: {minutes} minutes and {rest_seconds} seconds",term.Color.GREEN))
        print(term.format(f"Duty cycle: {'%.2f'%((1-gap_ratio)*100)}%", term.Color.GREEN))
        print(term.format(f"Total found frequencies: {total_f}", term.Color.GREEN))




# The cutoff ratio defines the ratio of the gap that is to be ignored.
cutoffRatio = 0.6
# How many frequencies will be looked at after the similar frequencies count
similarFrequenciesCount = 10
# How stark the deviation has to be to count as different within the last "similarFrequenciesCount" frequencies
similarityStdDev = 0.05
dieGracefully = False

maxiLength = None
minimumIntensity = None


from smurfs.support.singleton import Singleton
from enum import Enum

class UncertaintyMode(Enum):
    none = "None"
    montgomery = "Montgomery"
    fit = "Fit"

    @staticmethod
    def content():
        return [e.value for e in UncertaintyMode]


@Singleton
class Config:

    def __init__(self,*args):
        self.uncertaintiesMode = UncertaintyMode.none.value
        self.skipStdDevCheck = False
        self._skipSimilarFrequencies = False
        self.removeSector = []
        pass

    @property
    def uncertaintiesMode(self):
        return self._errorMode

    @uncertaintiesMode.setter
    def uncertaintiesMode(self, val : UncertaintyMode):
        if not isinstance(val, UncertaintyMode) and val not in UncertaintyMode.content():
            raise ValueError("Please pass a valid Errormode!")

        self._errorMode = val

    @property
    def skipSimilarFrequencies(self):
        return self._skipSimilarFrequencies

    @skipSimilarFrequencies.setter
    def skipSimilarFrequencies(self,val):
        self._skipSimilarFrequencies = val

    @property
    def skipCutoff(self):
        return self._skipCutoff

    @skipCutoff.setter
    def skipCutoff(self,val):
        self._skipCutoff = val

def conf() -> Config:
    return Config.ins()


import pip
from smurfs._version import __version__
pipVer = pip.__version__.split(".")

#main is not directly accessible anymore on pip > 10
if int(pipVer[0]) < 10:
    from pip import main
else:
    from pip._internal import main

from setuptools import setup,find_packages

reqs = ["astropy==3.0.5",
"matplotlib==3.0.2",
"numpy==1.15.4",
"pandas==0.23.4",
"pytest==3.10.1",
"scipy==1.1.0",
"stem==1.7.0",
"uncertainties==3.0.3",
"astroquery"]


#this is a bit of a hack, due to the inability to build numpy properly in a reasonable amount of time
for req in reqs:
        main(["install", req])

setup(
    name='smurfs',
    version=__version__,
    packages=find_packages(exclude=["*test","dist","build","venv"]),
    url='https://github.com/muma7490/SMURFS',
    license='MIT',
    author='Marco Müllner',
    author_email='muellnermarco@gmail.com',
    description='Smart UseR Frequency analySer, a fast and easy to use frequency analyser.',
)



import numpy as np
from smurfs.support import *
from typing import List,Dict,Tuple
from collections import OrderedDict
from uncertainties import unumpy
from astroquery.mast import Observations
from astropy.io import fits
import os
@timeit
def normalizeData(data: np.ndarray) -> np.ndarray:
    """
    This function reduces the dataset to make computations easier. It subtracts the 0 point in the temporal axis and
    removes the mean of the dataset.
    :param data: The dataset, probably read through readData
    :return: A normalized dataset
    """
    data[0] -= data[0][0]

    data = reduce_y(data)
    return data

def reduce_y(data: np.ndarray) -> np.ndarray:
    x = data[0]
    y = data[1]

    for i in ['nan','-nan']:
        if i in y.astype(str):
            n = len(y[y.astype(str)==i])
            print(term.format(f"You have {n} {i} in your data. These points will be removed.",term.Color.YELLOW))

        x = x[y.astype(str) != i]
        y = y[y.astype(str) != i]

    for i in [np.inf,-np.inf,np.nan]:
        if i in y:
            n = len(y[y==i])
            print(term.format(f"You have {n} {i} in your data. These points will be removed.",term.Color.YELLOW))

        x = x[y != i]
        y = y[y != i]

    y -= np.mean(y)
    return np.array((x, y))


@timeit
def readFrequencyMarker(file: str) -> Dict[str,float]:
    """
    Reads frequency Marker from file. First column needs to be the name of the marker, the second column needs to have
    the according frequency value for the marker.
    :param file: File where the marker are stored.
    :return: dict containing the dataset
    """
    if not isinstance(file,str):
        raise AttributeError("Parameter file is not a string! It is type {0}".format(type(file)))

    if not os.path.exists(file):
        raise FileNotFoundError("File {0} was not found!".format(file))

    marker = OrderedDict()
    with open(file,'r') as f:
        for line in f:
            try:
                name, value = str(line).split("\t")
            except ValueError:
                raise IOError("File {0} has not two ""columns! Needs to have only two. Has {1} "
                              "column".format(file,str(line).split("\t")))
            marker[name] = float(value)

    return marker
    pass

@timeit
def readData(file: str) -> np.ndarray:
    """
    This function checks if the filepath in file exists (if not it raises an IOError) and than reads the file into
    a numpy array
    :param file The filepath to the file which is to be read:
    :return Returns the array containing the data of file:
    """
    if not os.path.exists(file):
        raise IOError("File " + os.getcwd()+"/"+file + " doesn't exist!")

    return np.loadtxt(file).T

@timeit
def getSplits(data: np.ndarray, timeRange: float = -1, overlap: float = 0,ignoreCutoffRatio : bool = False) -> List[np.ndarray]:
    """
    The getSplits function creates equally sized chunks of the original dataset with an optional overlap between two
    chunks and returns a list of these arrays. To do this, it finds the nearest index to a datapoint which responds to
    the timeRange and slices the array accordingly.
    :param data: The dataset which is to be cut up
    :param timeRange:  The timeRange for one chunk. For example if overlap is 0 and timeRange is 50, the chunks will
    have the size of 0->50,50-100,...
    :param overlap: The overlap between two chunks. For example if overlap is 2 and timeRange is 50, the chunks will
    have the size of 0->50,48->95,...
    :return: A list of chunks of the original data
    """

    arg = np.argsort(data[0])
    data = np.array((data[0][arg],data[1][arg]))

    if data[0][0] != 0:
        data = normalizeData(data)

    if timeRange == -1:
        timeRange = max(data[0])
    dataPoints = []
    i = 0

    if not i*(timeRange-overlap)+timeRange <= max(data[0]):
        timeRange = max(data[0])

    while(i*(timeRange-overlap)+timeRange <= max(data[0])):
        lowerIndex = find_nearest(data[0],i*(timeRange-overlap))
        upperIndex = find_nearest(data[0],i*(timeRange-overlap)+timeRange)+1

        #this case can happen, if the full range is used, and no overlap is present
        if upperIndex == len(data[0]):
            upperIndex -= 1

        i += 1
        if timeRange != max(data[0]):
            gapRatio= getGapRatio(data,lowerIndex,upperIndex)
            if gapRatio > cutoffRatio:
                if not ignoreCutoffRatio:
                    print(term.format("Ignoring time base from "+str(int(data[0][lowerIndex]))+" to "+str(int(data[0][upperIndex]))+
                                      " because the gap Ratio is greater 0.5. Gap Ratio is "+str(gapRatio),term.Color.RED))
                    continue
                else:
                    print(term.format("Time base from  " + str(int(data[0][lowerIndex])) + " to " + str(
                        int(data[0][upperIndex])) +" has a gap ratio of  " + str(gapRatio) +
                          ", but will be taken eitherway because of ignoreCutoffRatio=True",term.Color.YELLOW))
        else:
            print(term.format("Ommited gap ratio. Time Range is "+str(timeRange)+",max data "+str(max(data[0])),term.Color.CYAN))
        array = np.array((data[0][lowerIndex:upperIndex],data[1][lowerIndex:upperIndex]))
        dataPoints.append(array)

    return dataPoints

def getMostCommonStep(data: np.ndarray):
    (values, counts) = np.unique(np.diff(data[0]), return_counts=True)
    ind = int(np.argmax(counts))
    if counts[ind]*100/np.sum(counts) < 10:
        counts[ind] = np.median(np.diff(data[0]))
    return ind,values

@timeit
def getGapRatio(data: np.ndarray, lowerIndex:int, upperIndex:int):
    """
    Calculates the ratio between the "empty" space within a given dataset to its datapoints. For example
    an array containing the datapoints 0,1,2,3,7,8,9 would return 0.3 as gap ratio, as one third of the data is not
    populated. To do this, it calculates the most common difference between two points, and sums up deviations for gaps
    bigger than this value.
    :param data: The dataset which needs to be analyzed
    :param lowerIndex: The lower index of the subset that is to be analyzed
    :param upperIndex: The upper index of the subset that is to be analyzed
    :return: A value between 0 and 1. The higher this value is, the worse the dataset is.
    """
    gaps = np.diff(data[0][lowerIndex:upperIndex])
    ind,values = getMostCommonStep(data)
    # adding 0.1, just to be sure not to get very small flukes in the observation time
    totalGap = np.sum(gaps[gaps>(1.2*values[ind])])

    return totalGap/(data[0][upperIndex]-data[0][lowerIndex])



@timeit
def find_nearest(array: np.ndarray, value: float) -> int:
    """
    Finds the index of the nearest value to a value in an array.
    :param array Array to be checked in:
    :param value value that needs to be found in array:
    :return index of nearest value:
    """
    idx = (np.abs(array-value)).argmin()
    return int(idx)

@timeit
def writeResults(file: str, data: Dict[Tuple[float,float],List[Tuple[float,float]]],nyquistFrequency : float, mode: str = 'w'):
    """
    This function writes the results of the analysis. The file will have a structure like this:
    Lowerrange Upperrange
    0 50
        frequency snr
        0.00    25
    :param file File where the results have to be written.
    :param data The dataset that is to be written. Needs to be a dict, where the key is a tuple containing min/max
    time points of the chunk, and the value a list of tuples, containing the frequencies and their corresponding snr
    :param mode write mode. Default is 'w':
    """

    for key,value in data.items():
        savePath = "{0:0=3d}".format(int(key[0]))
        savePath += "_{0:0=3d}/".format(int(key[1]))
        header = "f(c/d) f_err(c/d) snr amp amp_err phase phase_err residual_noise"
        arr = []
        for i in value:
            val_list = []
            for j in i:
                try:
                    val_list.append(j.nominal_value)
                    val_list.append(j.std_dev)
                except AttributeError:
                    val_list.append(j)
            arr.append(val_list)

        np.savetxt(savePath+file,np.array(arr),header=header,comments=f"Nyquist frequency(c/d):{nyquistFrequency}\n")

@timeit
def createPath(path):
    """
    This function creates a path if it doesn't exist
    :param path path to be created:
    """
    if not os.path.exists(path):
        os.makedirs(path)

def save_frequency_spacing(data : np.ndarray, path: str, name : str):
    createPath(path)
    with cd(path):
        try:
            val =  unumpy.nominal_values(data)
        except: # for uncertainties values
            val = data

        np.savetxt(f"{name}.txt", val.T)

        fig = pl.figure()
        pl.hist(val, bins='auto', color='k', rwidth=0.9)
        pl.xlabel(r"Frequency spacing ($d^{-1}$)")
        pl.ylabel("Counts")
        pl.title("Frequency spacing overview")
        fig.savefig(f"{name}.pdf")
        pl.close()



@timeit
def saveAmpSpectrumAndImage(ampSpectrum: np.ndarray, path: str, title : str,file_name: str):
    """
    This function saves the amplitude spectrum and according plots to the results path.
    :param name name of file and image:
    """
    createPath(path)

    with cd(path):
        np.savetxt(file_name+".txt",ampSpectrum.T)
        plotData = {title:(ampSpectrum, '-')}
        p = plotCustom(title,plotData,xLabel="Frequency(c/d)",yLabel="Amplitude")
        p.savefig(file_name+".pdf")
        pl.close()


def download_tic(tic_id : str):
    obsTable = Observations.query_criteria(dataproduct_type=["TIMESERIES"], obs_collection="TESS",target_name=tic_id)
    if len(tic_id) == 0:
        raise ValueError(f"Cannot find TESS data for {tic_id}")
    dataProductsByID = Observations.get_product_list(obsTable)
    Observations.download_products(dataProductsByID)
    fits_files = []
    for path,subfolder,files in os.walk("mastDownload"):
        tic_folders = []
        for folder in subfolder:
            if tic_id in folder:
                tic_folders.append(folder)

        if tic_folders == []:
            continue

        for folder in tic_folders:
            for _,_,files in os.walk(f"{path}/{folder}"):
                for i in files:
                    if i.endswith("lc.fits"):
                        fits_files.append(f"{path}/{folder}/{i}")

    if fits_files == []:
        raise ValueError(f"Couldn find proper TESS Data for {tic_id}")

    for file in fits_files:
        hdulist = fits.open(file)
        try:
            data_new = np.array((hdulist[1].data["TIME"], hdulist[1].data["PDCSAP_FLUX"]))
            data_new = reduce_y(data_new)
            data = np.hstack((data,data_new))
        except:
            data = np.array((hdulist[1].data["TIME"],hdulist[1].data["PDCSAP_FLUX"]))
            data = reduce_y(data)

    y = data[1] + 2*abs(min(data[1]))
    y = -2.5*np.log(y*2.5*10**(-18)/1000)
    y -= np.mean(y)
    data = np.array((data[0],y))

    return data
